/* macp.h - header file for MacPaint files
*/

#ifndef MACP_H_INCLUDED
#define MACP_H_INCLUDED

#define	HEADER_LENGTH	512
#define	MAX_LINES	720
#define	BYTES_WIDE	72
#define MAX_COLS	576	/* = BYTES_WIDE * 8 */

#endif
